import { BasePlugin } from '@opentelemetry/core';
import { SpanKind, Span, CanonicalCode } from '@opentelemetry/api';
import * as shimmer from 'shimmer';
import * as kafkaNode from 'kafka-node';
import { AttributeNames } from './enums';
import { Message, ProduceRequest } from 'kafka-node';
import { KafkaNodePluginConfig } from './types';
import { parseResponseStatus } from './utils';
import { VERSION } from './version';

export class KafkaNodePlugin extends BasePlugin<typeof kafkaNode> {

    protected _config!: KafkaNodePluginConfig;

    constructor(readonly moduleName: string) {
        super('@mzahor-test-org/open-telemetry-kafka-node', VERSION);
    }    

    protected patch() {
        this._logger.debug('kafka-node: patch kafka-node plugin');

        shimmer.wrap(this._moduleExports.Producer?.prototype, 'send', this._getProducerSendPatch.bind(this));
        shimmer.wrap(this._moduleExports.Consumer?.prototype, 'emit', this._getConsumerEmit.bind(this));

        return this._moduleExports;
    }

    protected unpatch() {
        this._logger.debug('kafka-node: unpatch kafka-node plugin');
        shimmer.unwrap(this._moduleExports.Producer?.prototype, 'send');
        shimmer.unwrap(this._moduleExports.Consumer?.prototype, 'emit');
    }

    private _getConsumerEmit(original: (event: string, ...args: unknown[]) => boolean) {
        const self = this;
        return function (event: string, ...args: unknown[]): boolean {

            if(event !== 'message') {
                return original.apply(this, arguments); 
            }

            const message = args[0] as Message;

            const brokers = Object.keys(this.client.brokers).join(',');
            const span = plugin._tracer.startSpan(message.topic, {
                kind: SpanKind.CONSUMER,
                attributes: {
                    [AttributeNames.MESSAGING_SYSTEM]: 'kafka',
                    [AttributeNames.MESSAGING_DESTINATION]: message.topic,
                    [AttributeNames.MESSAGING_DESTINATIONKIND]: 'topic',
                    [AttributeNames.NET_PEER_NAME]: brokers
                },
            });

            if(message.offset)
                span.setAttribute(AttributeNames.MESSAGING_OFFSET, message.offset);
            if(message.partition)
                span.setAttribute(AttributeNames.MESSAGING_PARTITION, message.partition);

            if(self._config?.consumerHook) {
                self._safeExecute([], () => self._config.consumerHook!(span, message), false);
            }

            try {
                return plugin._tracer.withSpan(span, () => {
                    return original.apply(this, arguments); 
                });
            }
            finally {
                span.setStatus({
                    code: CanonicalCode.OK,
                    message: 'OK'
                });
                span.end();
            }
        }
    }

    private _getProducerSendPatch(original: (payloads: ProduceRequest[], cb: (error: any, data: any) => any) => void) {

        const self = this;
        return function (payloads: ProduceRequest[], origCb: (error: any, data: any) => any) {
            
            const spans = payloads.flatMap( (payload: ProduceRequest) => {
                
                const brokers = Object.keys(this.client.brokers).join(',');
                let spansInThisRequest: Span[];
                if(Array.isArray(payload.messages)) {
                    spansInThisRequest = payload.messages.map(message => self._startProducerSpan(payload, message, brokers));
                }
                else {
                    spansInThisRequest = [self._startProducerSpan(payload, payload.messages, brokers)];    
                }
                
                return spansInThisRequest;
            });
            
            const patchedCb = (error: any, data: any): any => {
                const status = parseResponseStatus(error, data);
                self._safeExecute(spans, () => origCb(error, data), true);
                spans.forEach(span => { 
                    span.setStatus(status);
                    span.end();
                });
            }
        
            return original.call(this, payloads, patchedCb);
        }

    }

    private _startProducerSpan(producerRequest: ProduceRequest, message: any, brokers: string) {
        const span = plugin._tracer.startSpan(producerRequest.topic, {
            kind: SpanKind.PRODUCER,
            attributes: {
                [AttributeNames.MESSAGING_SYSTEM]: 'kafka',
                [AttributeNames.MESSAGING_DESTINATION]: producerRequest.topic,
                [AttributeNames.MESSAGING_DESTINATIONKIND]: 'topic',
                [AttributeNames.NET_PEER_NAME]: brokers
            },
        });

        if(this._config?.producerHook) {
            this._safeExecute([], () => this._config.producerHook!(span, producerRequest, message), false);
        }

        return span;
    }

    private _safeExecute<T extends (...args: unknown[]) => ReturnType<T>>(
        spans: Span[],
        execute: T,
        rethrow: boolean
      ): ReturnType<T> | void {
        try {
          return execute();
        } catch (error) {
          if (rethrow) {
              spans.forEach(span => {
                span.setStatus({
                    code: CanonicalCode.INTERNAL,
                    message: error?.message
                });
                span.end();    
              });
            throw error;
          }
          this._logger.error('caught error ', error);
        }
    }
    
    

}

export const plugin = new KafkaNodePlugin('kafka-node');