import { plugin } from '../src';
import { NoopLogger } from "@opentelemetry/core";
import { InMemorySpanExporter, SimpleSpanProcessor, ReadableSpan } from '@opentelemetry/tracing';
import { NodeTracerProvider } from '@opentelemetry/node';
import { AsyncHooksContextManager } from '@opentelemetry/context-async-hooks';
import { ContextManager } from '@opentelemetry/context-base';
import { context, SpanKind, CanonicalCode, Span } from '@opentelemetry/api';
import * as kafkaNode from 'kafka-node';
import { ProduceRequest, KafkaClient, Producer, Consumer, Message } from 'kafka-node';
import { AttributeNames } from '../src/enums';

describe('aspecto-open-telemetry-kafka-node', () => {

    const logger = new NoopLogger();
    const provider = new NodeTracerProvider();
    const memoryExporter = new InMemorySpanExporter();
    const spanProcessor = new SimpleSpanProcessor(memoryExporter);
    provider.addSpanProcessor(spanProcessor);
    let contextManager: AsyncHooksContextManager;

    let kafkaClient: KafkaClient;

    const getKafkaProducerSpans = (): ReadableSpan[] => {
        return memoryExporter
            .getFinishedSpans()
            .filter((s) => s.kind === SpanKind.PRODUCER);
    }

    const getKafkaConsumerSpans = (): ReadableSpan[] => {
        return memoryExporter
            .getFinishedSpans()
            .filter((s) => s.kind === SpanKind.CONSUMER);
    }

    beforeEach(() => {
        contextManager = new AsyncHooksContextManager();
        context.setGlobalContextManager(contextManager.enable() as unknown as ContextManager);

        this.kafkaClient = new KafkaClient({
            kafkaHost: 'testing_mock_host:1234'
        });    
    });

    afterEach(() => {
        memoryExporter.reset();
        contextManager.disable();
        this.kafkaClient.close();
        plugin.disable();
    });

    describe('successful producer.send', () => {

        beforeEach(() => {

            kafkaNode.Producer.prototype.send = function(payloads: ProduceRequest[], cb: (error: any, data: any) => any) {
                try{
                    cb(undefined, {});
                }
                catch(err) {}
            }

            plugin.enable(kafkaNode, provider, logger);
        });

        it('producer.send create span with right attributes', (done) => {
            const producer = new Producer(this.kafkaClient);

            const producerRequest = {
                topic: 'topic-name',
                messages: 'message content'
            };

            producer.send([producerRequest], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(1);
                    const span = spans[0];
                    expect(span.name).toStrictEqual('topic-name');
                    expect(span.status.code).toStrictEqual(CanonicalCode.OK);
                    expect(span.status.message).toStrictEqual('OK');
                    expect(span.attributes[AttributeNames.MESSAGING_SYSTEM]).toStrictEqual('kafka');
                    expect(span.attributes[AttributeNames.MESSAGING_DESTINATION]).toStrictEqual('topic-name');
                    expect(span.attributes[AttributeNames.MESSAGING_DESTINATIONKIND]).toStrictEqual('topic');
                    expect(span.attributes[AttributeNames.NET_PEER_NAME]).toStrictEqual('testing_mock_host:1234');

                    done();    
                })
            });                
        });

        it('two messages in request should create two spans', (done) => {
            const producer = new Producer(this.kafkaClient);

            const producerRequest = {
                topic: 'topic-name',
                messages: ['message content 1', 'message content 2']
            };

            producer.send([producerRequest], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(2);
                    spans.forEach(span => {
                        expect(span.name).toStrictEqual('topic-name');
                        expect(span.attributes[AttributeNames.MESSAGING_SYSTEM]).toStrictEqual('kafka');
                        expect(span.attributes[AttributeNames.MESSAGING_DESTINATION]).toStrictEqual('topic-name');
                        expect(span.attributes[AttributeNames.MESSAGING_DESTINATIONKIND]).toStrictEqual('topic');    
                    })

                    done();    
                })
            });                  
        });

        it('two requests in same send should create two spans', (done) => {
            const producer = new Producer(this.kafkaClient);

            const req1 = {
                topic: 'topic-name-1',
                messages: 'message content 1'
            };
            const req2 = {
                topic: 'topic-name-1',
                messages: 'message content 1'
            };

            producer.send([req1, req2], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(2);

                    const [s1, s2] = spans;
                    expect(s1.name).toStrictEqual(req1.topic);
                    expect(s1.attributes[AttributeNames.MESSAGING_DESTINATION]).toStrictEqual(req1.topic);
                    expect(s2.name).toStrictEqual(req2.topic);
                    expect(s2.attributes[AttributeNames.MESSAGING_DESTINATION]).toStrictEqual(req2.topic);

                    done();    
                })
            });                  
        });

        it('multiple requests and messages', (done) => {
            const producer = new Producer(this.kafkaClient);

            const req1 = {
                topic: 'topic-name-1',
                messages: ['message content 1', 'message content 2', 'message content 3']
            };
            const req2 = {
                topic: 'topic-name-1',
                messages: ['message content 1', 'message content 2']
            };

            producer.send([req1, req2], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(5);
                    done();    
                })
            });                  
        });

        it('send cb throw exception', (done) => {
            const producer = new Producer(this.kafkaClient);

            const req1 = {
                topic: 'topic-name-1',
                messages: 'message content 1'
            };

            producer.send([req1], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(1);
                    done();    
                });
                throw new Error('exception thrown in send callback');
            });                  
        });

    });

    describe('failure producer.send', () => {

        beforeEach(() => {

            kafkaNode.Producer.prototype.send = function(payloads: ProduceRequest[], cb: (error: any, data: any) => any) {
                setImmediate( () => cb(new Error('error message from cb'), undefined));
            }

            plugin.enable(kafkaNode, provider, logger);
        });

        it('error returned in cb sets the span status', done => {
            const producer = new Producer(this.kafkaClient);

            const req1 = {
                topic: 'topic-name-1',
                messages: 'message content 1'
            };

            producer.send([req1], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(1);
                    expect(spans[0].status.code).toBe(CanonicalCode.UNKNOWN);
                    expect(spans[0].status.message).toStrictEqual('error message from cb');
                    done();    
                })
            });                  
        });

    });

    describe('producer hook add attribute', () => {

        beforeEach(() => {

            kafkaNode.Producer.prototype.send = function(payloads: ProduceRequest[], cb: (error: any, data: any) => any) {
                setImmediate(() => cb(undefined, {}));
            }

            const config = {
                enabled: true,
                producerHook: (span: Span, producerRequest: ProduceRequest, message: any) => {
                    span.setAttribute('attribute from testing hook', message);
                }
            }
            plugin.enable(kafkaNode, provider, logger, config);
        });

        it('custom attributes from hook are added to span', done => {
            const producer = new Producer(this.kafkaClient);

            const req1 = {
                topic: 'topic-name-1',
                messages: ['message content 1', 'message content 2']
            };

            producer.send([req1], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(2);
                    const [s1, s2] = spans;
                    expect(s1.attributes['attribute from testing hook']).toStrictEqual('message content 1');
                    expect(s2.attributes['attribute from testing hook']).toStrictEqual('message content 2');
                    done();    
                })
            });                  
        });

    });

    describe('producer hook throws exception', () => {

        beforeEach(() => {

            kafkaNode.Producer.prototype.send = function(payloads: ProduceRequest[], cb: (error: any, data: any) => any) {
                setImmediate(() => cb(undefined, {}));
            }

            const config = {
                enabled: true,
                producerHook: (span: Span, producerRequest: ProduceRequest, message: any) => {
                    throw Error('exception thrown from producer hook')
                }
            }
            plugin.enable(kafkaNode, provider, logger, config);
        });

        it('spans should be created when exception thrown in hook', done => {
            const producer = new Producer(this.kafkaClient);

            const req1 = {
                topic: 'topic-name-1',
                messages: ['message content 1', 'message content 2']
            };

            producer.send([req1], (error, data) => {
                setImmediate( () => {
                    const spans = getKafkaProducerSpans();
                    expect(spans.length).toBe(2);
                    expect(spans[0].status.code).toBe(CanonicalCode.OK);
                    done();    
                })
            });                  
        });

    });

    describe('kafka consumer', () => {

        beforeEach(() => {
            plugin.enable(kafkaNode, provider, logger);
        });

        it('consumer create span with right attributes', () => {
            const consumer = new Consumer(this.kafkaClient, [], {});
            const msg: Message = {
                topic: 'topic-name',
                value: 'content of message',
                offset: 1234,
                partition: 4
            }
            consumer.emit('message', msg);
            const spans = getKafkaConsumerSpans();
            expect(spans.length).toBe(1);
            const span = spans[0];
            expect(span.name).toStrictEqual(msg.topic);
            expect(span.status.code).toStrictEqual(CanonicalCode.OK);
            expect(span.status.message).toStrictEqual('OK');
            expect(span.attributes[AttributeNames.MESSAGING_SYSTEM]).toStrictEqual('kafka');
            expect(span.attributes[AttributeNames.MESSAGING_DESTINATION]).toStrictEqual(msg.topic);
            expect(span.attributes[AttributeNames.MESSAGING_DESTINATIONKIND]).toStrictEqual('topic');
            expect(span.attributes[AttributeNames.MESSAGING_PARTITION]).toStrictEqual(msg.partition);
            expect(span.attributes[AttributeNames.MESSAGING_OFFSET]).toStrictEqual(msg.offset);
            expect(span.attributes[AttributeNames.NET_PEER_NAME]).toStrictEqual('testing_mock_host:1234');
        });

        it('data missing in message should not create attribute', () => {
            const consumer = new Consumer(this.kafkaClient, [], {});
            const msg: Message = {
                topic: 'topic-name',
                value: 'content of message',
                // missing partition and offset
            }
            consumer.emit('message', msg);
            const spans = getKafkaConsumerSpans();
            expect(spans.length).toBe(1);
            const span = spans[0];
            expect(span.attributes.hasOwnProperty(AttributeNames.MESSAGING_PARTITION)).toBeFalsy();
            expect(span.attributes.hasOwnProperty(AttributeNames.MESSAGING_OFFSET)).toBeFalsy();
        });

        it('ignoring events not of type `message`', () => {
            const consumer = new Consumer(this.kafkaClient, [], {});
            consumer.emit('some other event', new Error('some error which we dont care about'));
            const spans = getKafkaConsumerSpans();
            expect(spans.length).toBe(0);
        });
    });

    describe('kafka consumer custom hook', () => {

        beforeEach(() => {
            const config = {
                enabled: true,
                consumerHook: (span: Span, message: Message) => {
                    span.setAttribute('attribute from testing hook', message.value);
                }
            }
            plugin.enable(kafkaNode, provider, logger, config);
        });

        it('attribute from consumer hook should be set on span', () => {
            const consumer = new Consumer(this.kafkaClient, [], {});
            const msg: Message = {
                topic: 'topic-name',
                value: 'content of message',
            }
            consumer.emit('message', msg);
            const spans = getKafkaConsumerSpans();
            expect(spans.length).toBe(1);
            const span = spans[0];
            expect(span.attributes['attribute from testing hook']).toStrictEqual(msg.value);
        });

    });

    describe('kafka consumer custom hook throws', () => {

        beforeEach(() => {
            const config = {
                enabled: true,
                consumerHook: (span: Span, message: Message) => {
                    throw Error('exception thrown from consumer hook')
                }
            }
            plugin.enable(kafkaNode, provider, logger, config);
        });

        it('spans should be created when exception thrown in hook', () => {
            const consumer = new Consumer(this.kafkaClient, [], {});
            const msg: Message = {
                topic: 'topic-name',
                value: 'content of message',
            }
            consumer.emit('message', msg);
            const spans = getKafkaConsumerSpans();
            expect(spans.length).toBe(1);
            const span = spans[0];
            expect(span.status.code).toStrictEqual(CanonicalCode.OK);
        });

    });

});